from __future__ import print_function

import base64
import json
from shutil import which
from time import sleep
import logging

from blessings import Terminal
from kubernetes import client, config
from kubernetes.client.rest import ApiException

from nephos.helpers.misc import execute, input_files, pretty_print

TERM = Terminal()


# Configs can be set in Configuration class directly or using helper utility
if which("kubectl"):
    config.load_kube_config()
    api = client.CoreV1Api()
    api_ext = client.ExtensionsV1beta1Api()
else:  # pragma: no cover
    logging.critical(TERM.red('We do not have "kubectl" installed'))


# Class to execute K8S commands
# TODO: We might wish to set the container at the execution level?
class Executer:
    def __init__(self, pod, namespace, container=""):
        """Executer creates a K8S pod object capable of:
        1) Execute commands,
        2) Return logs.

        Args:
            pod (str): Pod to bind to.
            namespace (str): Name of namespace.
            container (str): Container to bind to.
        """
        extra = ""
        if container:
            extra += f"--container {container} "
        self.pod = pod
        self.prefix_exec = f"kubectl exec {pod} -n {namespace} {extra}-- "

        self.prefix_logs = f"kubectl logs {pod} -n {namespace} {extra}"

    # TODO: api.connect_get_namespaced_pod_exec (to do exec using Python API programmatically)
    def execute(self, command):
        """Execute a command in pod.

        Args:
            command (str): Command to execute.

        Returns:
            tuple: 2-tuple of execution info:
            1) result of the command, if successful, None if not;
            2) and error, if command failed, None if not.

        """
        result, error = execute(self.prefix_exec + command)
        return result, error

    def logs(self, tail=-1, since_time=None):
        """Get logs from pod.

        Args:
            tail (int): How many lines of logs to obtain?

        Returns:
            str: Logs contained in pod.
        """
        command = f"--tail={tail}"
        if since_time:
            command += f" --since-time='{since_time}'"
        result, _ = execute(self.prefix_logs + command)
        return result


# Config
def context_get():
    """Obtain active K8S context.

    Returns:
        object: Active context.
    """
    _, active_context = config.list_kube_config_contexts()
    logging.debug(pretty_print(json.dumps(active_context)))
    return active_context


# Namespaces
def ns_create(namespace):
    """Create K8S namespace.

    Args:
        namespace (str): Name of namespace.
    """
    try:
        ns_read(namespace)
    except ApiException:
        ns = client.V1Namespace()
        ns.metadata = client.V1ObjectMeta(name=namespace)
        api.create_namespace(ns)
        logging.info(f'Created namespace "{namespace}"')
        logging.debug(pretty_print(json.dumps(ns.metadata, default=str)))


# TODO: Can we be more precise with the return type annotation?
def ns_read(namespace):
    """Read Name of namespace.

    Args:
        namespace (str): Name of namespace.

    Returns:
        object: Namespace object.
    """
    ns = api.read_namespace(name=namespace)
    logging.debug(pretty_print(json.dumps(ns.metadata, default=str)))
    return ns


# Ingress
# TODO: Convert list to tuple
def ingress_read(name, namespace="default"):
    """Get host names contained in K8S Ingress.

    Args:
        name (str): Name of Ingress.
        namespace (str): Name of namespace.

    Returns:
        list: List of host names.
    """
    ingress = api_ext.read_namespaced_ingress(name=name, namespace=namespace)
    hosts = [item.host for item in ingress.spec.rules if item.host]
    logging.debug(pretty_print(json.dumps(hosts)))
    return hosts


# Pods
# TODO: We should not need to specify pod number.
def pod_check(namespace, identifier, sleep_interval=10, pod_num=None):
    """Check if a set of pods exist and are functional.

    Args:
        namespace (str): Namespace where Helm deployment is located.
        identifier (str): Name of pod, or a label descriptor.
        sleep_interval (int): Number of seconds to sleep between attempts.
        pod_num (int): Number of pods expected to exist in the release. None by default.
    """
    logging.info("Ensuring that all pods are running ")
    running = False
    while not running:
        states, _ = execute(
            f'kubectl get pods -n {namespace} {identifier} -o jsonpath="{{.items[*].status.phase}}"'
        )
        states_list = states.split()
        # Let us also check the number of pods we have
        # We keep checking the state of the pods until they are running
        states = set(states_list)
        if (
            len(states) == 1
            and "Running" in states
            and (pod_num is None or len(states_list) == pod_num)
        ):
            logging.info(TERM.green("All pods are running"))
            running = True
        else:
            print(TERM.red("."), end="", flush=True)
            sleep(sleep_interval)


# Configmaps and secrets
# TODO: Refactor these so we have the same API as with secrets
def cm_create(cm_data, name, namespace="default"):
    """Create a K8S ConfigMap

    Args:
        cm_data (dict): Data to store in ConfigMap as key/value hash.
        name (str): Name of ConfigMap.
        namespace (str): Name of namespace.
    """
    # TODO: We should check that CM exists before we create it
    cm = client.V1ConfigMap()
    cm.metadata = client.V1ObjectMeta(name=name)
    cm.data = cm_data
    api.create_namespaced_config_map(namespace=namespace, body=cm)
    logging.info(f"Created ConfigMap {name} in namespace {namespace}")


def cm_read(name, namespace="default"):
    """Read a K8S ConfigMap.

    Args:
        name (str): Name of the ConfigMap.
        namespace (str): Name of namespace.

    Returns:
        dict: Keys and values stored in the ConfigMap.
    """
    cm = api.read_namespaced_config_map(name=name, namespace=namespace)
    logging.debug(pretty_print(json.dumps(cm.data)))
    return cm.data


def secret_create(secret_data, name, namespace="default"):
    """Create a K8S Secret.

    Args:
        secret_data (dict): Data to store in t as key/value hash.
        name (str): Name of the Secret.
        namespace (str): Name of namespace.
    """
    # Encode the data in a copy of the input dictionary
    # TODO: We should check that Secret exists before we create it
    secret_data = secret_data.copy()
    for key, value in secret_data.items():
        if isinstance(value, str):
            value = value.encode("ascii")
        secret_data[key] = base64.b64encode(value).decode("utf-8")
    secret = client.V1Secret()
    secret.metadata = client.V1ObjectMeta(name=name)
    secret.type = "Opaque"
    secret.data = secret_data
    api.create_namespaced_secret(namespace=namespace, body=secret)
    logging.info(f"Created Secret {name} in namespace {namespace}")


def secret_read(name, namespace="default"):
    """Read a K8S Secret.

    Args:
        name (str): Name of the Secret.
        namespace (str): Name of namespace.

    Returns:
        dict: Keys and values stored in the Secret.
    """
    secret = api.read_namespaced_secret(name=name, namespace=namespace)
    for key, value in secret.data.items():
        if value:
            secret.data[key] = base64.b64decode(value).decode("utf-8", "ignore")
    logging.debug(pretty_print(json.dumps(secret.data)))
    return secret.data


def secret_from_file(secret, namespace, key=None, filename=None):
    """Convert a file into a K8S Secret.

    Args:
        secret (str): Name of Secret where to save the file.
        namespace (str): Name of namespace.
        key (str): Key to which to assign the file in the K8S t. If not specified, the filename is used.
        filename (str): If not provided, we ask the user for input.
    """
    try:
        secret_read(secret, namespace)
    except ApiException:
        # Get relevant variables
        if not filename:
            secret_data = input_files((key,), clean_key=True)
        else:
            with open(filename, "rb") as f:
                data = f.read()
                secret_data = {key: data}
        secret_create(secret_data, secret, namespace)


def get_app_info(namespace, ingress, secret, secret_key="API_KEY"):
    """Get application information.

    Args:
        namespace (str): Name of namespace.
        ingress (str): Ingress name.
        secret (str): Secret where access details (e.g. API key) are located.
        secret_key (str): Key in t containing access details. By default "API KEY"

    Returns:

    """
    # Get ingress URL
    ingress_data = ingress_read(ingress, namespace=namespace)
    url = ingress_data[0]
    # Get API_KEY from secret
    secret_data = secret_read(secret, namespace)
    apikey = secret_data[secret_key]
    # Return data
    data = {"api-key": apikey, "url": url}
    return data



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from os import path
from time import sleep
import logging

from kubernetes.client.rest import ApiException
from nephos.fabric.settings import get_namespace, get_version
from nephos.fabric.utils import get_helm_pod
from nephos.helpers.helm import (
    HelmPreserve,
    helm_check,
    helm_extra_vars,
    helm_install,
    helm_upgrade,
)
from nephos.helpers.k8s import ingress_read, secret_read
from nephos.helpers.misc import execute_until_success

CURRENT_DIR = path.abspath(path.split(__file__)[0])


# Core sub-functions
def ca_chart(opts, release, upgrade=False):
    """Deploy CA Helm chart to K8S.

    Args:
        opts (dict): Nephos options dict.
        release (str): Name of the Helm Chart release.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    values_dir = opts["core"]["dir_values"]
    repository = opts["core"]["chart_repo"]
    ca_namespace = get_namespace(opts, ca=release)
    # PostgreSQL (Upgrades here are dangerous, deactivated by default)
    # Upgrading database is risky, so we disallow it by default
    if not upgrade:
        version = get_version(opts, "postgresql")
        config_yaml = f"{values_dir}/postgres-ca/{release}-pg.yaml"
        extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
        helm_install(
            "stable",
            "postgresql",
            f"{release}-pg",
            ca_namespace,
            extra_vars=extra_vars,
            
        )
        helm_check("postgresql", f"{release}-pg", ca_namespace)
    psql_secret = secret_read(
        f"{release}-pg-postgresql", ca_namespace
    )
    # Different key depending of PostgreSQL version
    psql_password = (
        psql_secret.get("postgres-password") or psql_secret["postgresql-password"]
    )
    # Fabric CA
    version = get_version(opts, "hlf-ca")
    env_vars = [("externalDatabase.password", psql_password)]
    config_yaml = f"{values_dir}/hlf-ca/{release}.yaml"
    if not upgrade:
        extra_vars = helm_extra_vars(
            version=version, config_yaml=config_yaml, env_vars=env_vars
        )
        helm_install(
            repository,
            "hlf-ca",
            release,
            ca_namespace,
            extra_vars=extra_vars,
            
        )
    else:
        preserve = (
            HelmPreserve(
                ca_namespace,
                f"{release}-hlf-ca--ca",
                "CA_ADMIN",
                "adminUsername",
            ),
            HelmPreserve(
                ca_namespace,
                f"{release}-hlf-ca--ca",
                "CA_PASSWORD",
                "adminPassword",
            ),
        )
        extra_vars = helm_extra_vars(
            version=version,
            config_yaml=config_yaml,
            env_vars=env_vars,
            preserve=preserve,
        )
        helm_upgrade(
            repository, "hlf-ca", release, extra_vars=extra_vars
        )
    helm_check("hlf-ca", release, ca_namespace)


def ca_enroll(pod_exec):
    """Enroll CA.

    Enroll the Certificate Authority (CA) identity within the running CA pod.
    This is a necessary step for the CA to function.

    Args:
        pod_exec: A pod executor instance bound to the CA.
    """
    alive = False
    while not alive:
        res = pod_exec.logs()
        if "Listening on" in res:
            alive = True
        else:
            sleep(15)
    # Enroll CA Admin if necessary

    # TODO: Add verification for pem file.
    ca_cert, _ = pod_exec.execute(
        "cat /var/hyperledger/fabric-ca/msp/signcerts/cert.pem"
    )
    if not ca_cert:
        enrolled_id = False
        while not enrolled_id:
            res, err = pod_exec.execute(
                "bash -c 'fabric-ca-client enroll -d -u http://$CA_ADMIN:$CA_PASSWORD@$SERVICE_DNS:7054'"
            )
            if not err:
                enrolled_id = True
            else:
                sleep(15)


def check_ca(ingress_host, cacert=None):
    """Check that the CA Ingress is responsive.

    Args:
        ingress_host (str): Ingress host for the CA.
        cacert (str): Path of the CA cert.
        
    """
    # Check that CA ingress is operational
    command = f"curl https://{ingress_host}/cainfo"
    if cacert:
        command += f" --cacert {cacert}"
    execute_until_success(command)


# Runner
def setup_ca(opts, upgrade=False):
    """Setup CA.

    Setup involves enrolling the CA admin, checking the Ingress
    is responsive.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    for ca_name, ca_values in opts["cas"].items():
        ca_namespace = get_namespace(opts, ca=ca_name)
        # Install Charts
        ca_chart(opts=opts, release=ca_name, upgrade=upgrade)

        # Obtain CA pod and Enroll
        pod_exec = get_helm_pod(
            namespace=ca_namespace, release=ca_name, app="hlf-ca"
        )
        ca_enroll(pod_exec)

        # Get CA Ingress and check it is running
        try:
            # Get ingress of CA
            ingress_urls = ingress_read(
                ca_name + "-hlf-ca", namespace=ca_namespace
            )
        except ApiException:
            logging.warning("No ingress found for CA")
            continue

        # Check the CA is running
        check_ca(
            ingress_host=ingress_urls[0],
            cacert=ca_values.get("tls_cert"),
            
        )



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import shutil
from collections import namedtuple
from glob import glob
from os import chdir, getcwd, listdir, makedirs
from os.path import abspath, exists, isfile, isdir, join, split
from time import sleep
import logging

from nephos.fabric.settings import get_namespace
from nephos.helpers.k8s import ns_create, ingress_read, secret_from_file
from nephos.helpers.misc import execute, execute_until_success
from nephos.fabric.utils import (
    get_msps,
    get_channels,
    get_peers,
    get_orderers,
    is_orderer_msp,
    credentials_secret,
    crypto_secret,
    get_helm_pod,
    get_secret_genesis
)

PWD = getcwd()
CryptoInfo = namedtuple("CryptoInfo", ("secret_type", "subfolder", "key", "required"))


# CA Helpers
def check_id(ca_namespace, ca, username):
    """

    Args:
        ca_namespace (str): K8S namespace where CA is located.
        ca (str): K8S release name of CA.
        username (str): Username for identity.

    Returns:
        bool: Does the ID exist?
    """
    # Get CA
    ca_exec = get_helm_pod(
        namespace=ca_namespace, release=ca, app="hlf-ca"
    )
    # Check if Orderer is registered with the relevant CA
    got_id = False
    while not got_id:
        ord_id, err = ca_exec.execute(
            f"fabric-ca-client identity list --id {username}"
        )
        if err:
            # Expected error (identity does not exist)
            if "no rows in result set" in err:
                got_id = True
            # Otherwise, unexpected error, we are having issues connecting to CA
            else:
                sleep(15)
        else:
            got_id = True
    return ord_id


def register_id(
    ca_namespace, ca, username, password, node_type="client", admin=False):
    """Register an ID with a Fabric Certificate Authority

    Args:
        ca_namespace (str): K8S namespace where CA is located.
        ca (str): K8S release name of CA.
        username (str): Username for identity.
        password (str): Password for identity.
        node_type (str): Node type for identity. "client" by default.
        admin (bool): Whether the identity is an admin. False by default.
    """
    # Get CA
    ord_id = check_id(ca_namespace, ca, username)
    # Registered if needed
    ca_exec = get_helm_pod(namespace=ca_namespace, release=ca, app="hlf-ca")
    if not ord_id:
        command = (
            "fabric-ca-client register --id.name {id} --id.secret {pw} --id.type {type}"
        )
        if admin:
            command += " --id.attrs 'admin=true:ecert'"
        registered_id = False
        while not registered_id:
            res, err = ca_exec.execute(
                command.format(id=username, pw=password, type=node_type)
            )
            if not err:
                registered_id = True
            # Otherwise, unexpected error, we are having issues connecting to CA
            else:
                sleep(15)


def enroll_id(opts, ca, username, password):
    """Enroll an ID with a Fabric Certificate Authority

    Args:
        opts (dict): Nephos options dict.
        ca (str): K8S release name of CA.
        username (str): Username for identity.
        password (str): Password for identity.

    Returns:
        str: Path of the MSP directory where cryptographic data is saved.

    """
    dir_crypto = opts["core"]["dir_crypto"]
    ca_namespace = get_namespace(opts, ca=ca)
    ingress_urls = ingress_read(ca + "-hlf-ca", namespace=ca_namespace)
    msp_dir = f"{username}_MSP"
    msp_path = join(dir_crypto, msp_dir)
    if not isdir(msp_path):
        # Enroll
        command = (
            f'FABRIC_CA_CLIENT_HOME={dir_crypto} fabric-ca-client enroll '
            + f'-u https://{username}:{password}@{ingress_urls[0]} -M {join(dir_crypto, msp_dir)} '
            + f'--tls.certfiles {abspath(opts["cas"][ca]["tls_cert"])}'
        )
        execute_until_success(command)
    return msp_path


def create_admin(opts, msp_name):
    """Create an admin identity.

    Args:
        opts (dict): Nephos options dict.
        msp_name (str): Name of Membership Service Provider.
    """
    dir_config = opts["core"]["dir_config"]
    dir_crypto = opts["core"]["dir_crypto"]
    msp_values = opts["msps"][msp_name]
    ca_values = opts["cas"][msp_values["ca"]]

    # TODO: Refactor this into its own function
    ca_name = msp_values["ca"]
    ca_namespace = get_namespace(opts, ca=ca_name)

    # Get CA ingress
    ingress_urls = ingress_read(
        ca_name + "-hlf-ca", namespace=ca_namespace)
    ca_ingress = ingress_urls[0]

    # Register the Organisation with the CAs
    register_id(
        ca_namespace,
        msp_values["ca"],
        msp_values["org_admin"],
        msp_values["org_adminpw"],
        admin=True,
    )

    # TODO: Can we reuse the Enroll function above?
    # If our keystore does not exist or is empty, we need to enroll the identity...
    keystore = join(dir_crypto, msp_name, "keystore")
    if not isdir(keystore) or not listdir(keystore):
        execute(
            (
                f"FABRIC_CA_CLIENT_HOME={dir_config} fabric-ca-client enroll "
                + f"-u https://{msp_values['org_admin']}:{msp_values['org_adminpw']}@{ca_ingress} "
                + f"-M {join(dir_crypto, msp_name)} --tls.certfiles {abspath(ca_values['tls_cert'])}"
            ),
        )


def admin_creds(opts, msp_name):
    """Get admin credentials and save them to Nephos options dict.

    Args:
        opts (dict): Nephos options dict.
        msp_name (str): Name of Membership Service Provider.
    """
    msp_namespace = get_namespace(opts, msp=msp_name)
    msp_values = opts["msps"][msp_name]

    admin_cred_secret = f"hlf--{msp_values['org_admin']}-admincred"
    secret_data = credentials_secret(
        admin_cred_secret,
        msp_namespace,
        username=msp_values["org_admin"],
        password=msp_values.get("org_adminpw"),
    )
    msp_values["org_adminpw"] = secret_data["CA_PASSWORD"]


# TODO: Rename to something more appropriate (e.g. copy_msp_file)
def copy_secret(from_dir, to_dir):
    """Copy single secret file from one directory to another.

    Args:
        from_dir (str): Source directory where file resides.
        to_dir (str): Destination directory for file.
    """
    from_list = glob(join(from_dir, "*"))
    if len(from_list) == 1:
        from_file = from_list[0]
    else:
        raise ValueError(
            f"from_dir contains {len(from_list)} files - {from_list}"
        )
    _, from_filename = split(from_file)
    to_file = join(to_dir, from_filename)
    if not isfile(to_file):
        if not isdir(to_dir):
            makedirs(to_dir)
        shutil.copy(from_file, to_file)


def msp_secrets(opts, msp_name):
    """Process MSP and convert it to a set of secrets.

    Args:
        opts (dict): Nephos options dict.
        msp_name (str): Name of Membership Service Provider.
    """
    # Relevant variables
    msp_namespace = get_namespace(opts, msp=msp_name)
    msp_values = opts["msps"][msp_name]
    if opts["cas"]:
        # If we have a CA, MSP was saved to dir_crypto
        msp_path = join(opts["core"]["dir_crypto"], msp_name)
    else:
        # Otherwise we are using Cryptogen
        glob_target = f"{opts['core']['dir_crypto']}/crypto-config/*Organizations/{msp_namespace}*/users/Admin*/msp"
        msp_path_list = glob(glob_target)
        if len(msp_path_list) == 1:
            msp_path = msp_path_list[0]
        else:
            raise ValueError(
                f"MSP path list length is {msp_path_list} - {msp_path_list}"
            )

    # Copy cert to admincerts
    copy_secret(join(msp_path, "signcerts"), join(msp_path, "admincerts"))

    # Create ID secrets from Admin MSP
    id_to_secrets(msp_namespace, msp_path, msp_values["org_admin"])

    # Create CA secrets from Admin MSP
    cacerts_to_secrets(msp_namespace, msp_path, msp_values["org_admin"])


def admin_msp(opts, msp_name):
    """Setup the admin MSP, by getting/setting credentials and creating/saving crypto-material.

    Args:
        opts (dict): Nephos options dict.
        msp_name (str): Name of Membership Service Provider.
    """
    admin_namespace = get_namespace(opts, msp_name)
    ns_create(admin_namespace)

    if opts["cas"]:
        # Get/set credentials (if we use a CA)
        admin_creds(opts, msp_name)
        # Crypto material for Admin
        create_admin(opts, msp_name)
    else:
        logging.info("No CAs defined in Nephos settings, ignoring Credentials")

    # Setup MSP secrets
    msp_secrets(opts, msp_name)


# General helpers
def item_to_secret(namespace, msp_path, username, item):
    """Save a single MSP crypto-material file as a K8S secret.

    Args:
        namespace (str): Namespace where secret will live.
        msp_path (str): Path to the Membership Service Provider crypto-material.
        username (str): Username for identity.
        item (CryptoInfo): Item containing cryptographic material information.
    """
    # Item in form CryptoInfo(name, subfolder, key, required)
    secret_name = f"hlf--{username}-{item.secret_type}"
    file_path = join(msp_path, item.subfolder)
    try:
        crypto_secret(secret_name, namespace, file_path=file_path, key=item.key)
    except Exception as error:
        if item.required:
            raise Exception(error)
        else:
            logging.warning(
                f'No {file_path} found, so secret "{secret_name}" was not created'
            )


def id_to_secrets(namespace, msp_path, username):
    """Convert Identity certificate and key to K8S secrets.

    Args:
        namespace (str): Namespace where secret will live.
        msp_path (str): Path to the Membership Service Provider crypto-material.
        username (str): Username for identity.
    """
    crypto_info = [
        CryptoInfo("idcert", "signcerts", "cert.pem", True),
        CryptoInfo("idkey", "keystore", "key.pem", True),
    ]
    for item in crypto_info:
        item_to_secret(namespace, msp_path, username, item)


def cacerts_to_secrets(namespace, msp_path, user):
    """Convert CA certificate to K8S secrets.

    Args:
        namespace (str): Namespace where secret will live.
        msp_path (str): Path to the Membership Service Provider crypto-material.
        username (str): Username for identity.
    """
    crypto_info = [
        CryptoInfo("cacert", "cacerts", "cacert.pem", True),
        CryptoInfo("caintcert", "intermediatecerts", "intermediatecacert.pem", False),
    ]
    for item in crypto_info:
        item_to_secret(namespace, msp_path, user, item)


def setup_id(opts, msp_name, release, id_type):
    """Setup single ID by registering, enrolling, and saving ID to K8S secrets.

    Args:
        opts (dict): Nephos options dict.
        msp_name (str): Name of Membership Service Provider.
        release (str): Name of release/node.
        id_type (str): Type of ID we use.
    """
    msp_values = opts["msps"][msp_name]
    node_namespace = get_namespace(opts, msp_name)
    if opts["cas"]:
        ca_namespace = get_namespace(opts, ca=opts["msps"][msp_name]["ca"])
        # Create secret with Orderer credentials
        secret_name = f"hlf--{release}-cred"
        secret_data = credentials_secret(
            secret_name, node_namespace, username=release
        )
        # Register node
        register_id(
            ca_namespace,
            msp_values["ca"],
            secret_data["CA_USERNAME"],
            secret_data["CA_PASSWORD"],
            id_type,
        )
        # Enroll node
        msp_path = enroll_id(
            opts,
            msp_values["ca"],
            secret_data["CA_USERNAME"],
            secret_data["CA_PASSWORD"],
        )
    else:
        # Otherwise we are using Cryptogen
        glob_target = f"{opts['core']['dir_crypto']}/crypto-config/{id_type}Organizations/{node_namespace}*/{id_type}s/{release}*/msp"
        msp_path_list = glob(glob_target)
        if len(msp_path_list) == 1:
            msp_path = msp_path_list[0]
        else:
            raise ValueError(
                f"MSP path list length is {msp_path_list} - {msp_path_list}"
            )
    # Secrets
    id_to_secrets(
        namespace=node_namespace, msp_path=msp_path, username=release
    )


# TODO: Rename to mention identities.
def setup_nodes(opts):
    """Setup identities for nodes.

    Args:
        opts (dict): Nephos options dict.
    """
    for msp in get_msps(opts=opts):
        for peer in get_peers(opts=opts, msp=msp):
            setup_id(opts, msp, peer, "peer")

    for msp in get_msps(opts=opts):
        for orderer in get_orderers(opts=opts, msp=msp):
            setup_id(opts, msp, orderer, "orderer")


# ConfigTxGen helpers
def genesis_block(opts):
    """Create and save Genesis Block to K8S.

    Args:
        opts (dict): Nephos options dict.
    """
    # Change to blockchain materials directory
    chdir(opts["core"]["dir_config"])
    # Create the genesis block
    genesis_key = "genesis.block"
    genesis_file = join(opts["core"]["dir_crypto"], genesis_key)
    if not exists(genesis_file):
        # Genesis block creation and storage
        execute(
            f"configtxgen -profile OrdererGenesis -outputBlock {genesis_file}",
        )
    else:
        logging.info(f"{genesis_file} already exists")

    for msp in get_msps(opts=opts):
        if not is_orderer_msp(opts=opts, msp=msp):
            continue
        ord_namespace = get_namespace(opts, msp=msp)
        # Create the genesis block secret
        secret_from_file(
            secret=get_secret_genesis(opts=opts),
            namespace=ord_namespace,
            key=genesis_key,
            filename=genesis_file,
        )
        # Return to original directory
    chdir(PWD)


def channel_tx(opts):
    """Create and save Channel Transaction to K8S.

    Args:
        opts (dict): Nephos options dict.
    """
    # Change to blockchain materials directory
    chdir(opts["core"]["dir_config"])
    # Create Channel Tx
    for channel in get_channels(opts):
        channel_key = f"{opts['channels'][channel]['channel_name']}.tx"
        channel_file = join(opts["core"]["dir_crypto"], channel_key)
        if not exists(channel_file):
            # Channel transaction creation and storage
            execute(
                f"configtxgen -profile {opts['channels'][channel]['channel_profile']} -channelID {opts['channels'][channel]['channel_name']} -outputCreateChannelTx {channel_file}",
            )
        else:
            logging.info(f"{channel_file} already exists")
        # Create the channel transaction secret
        for msp in get_msps(opts=opts):
            if msp not in opts["channels"][channel]["msps"]:
                continue
            peer_namespace = get_namespace(opts, msp=msp)
            secret_from_file(
                secret=opts['channels'][channel]["secret_channel"],
                namespace=peer_namespace,
                key=channel_key,
                filename=channel_file,
            )
            # Return to original directory
    chdir(PWD)



#! /usr/bin/env python

#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
import logging
import click
from blessings import Terminal

from nephos.runners import (
    runner_ca,
    runner_composer,
    runner_composer_up,
    runner_crypto,
    runner_deploy,
    runner_fabric,
    runner_orderer,
    runner_peer,
)

from nephos.fabric.settings import load_config
from nephos.helpers.misc import pretty_print

TERM = Terminal()
log_format = '%(asctime)s %(module)-10s %(levelname)-8s %(message)s'
logging_level = logging.INFO
log_file = "/tmp/nephos_log"

class Settings(object):
    def __init__(self, settings_file, upgrade, verbose, out):
        self.settings_file = settings_file
        self.upgrade = upgrade
        self.verbose = verbose
        self.out = out


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group(
    help=TERM.green("Nephos helps you install Hyperledger Fabric on Kubernetes")
)
@click.option(
    "--settings_file",
    "-f",
    required=True,
    help=TERM.cyan("YAML file containing HLF options"),
)
@click.option(
    "--upgrade",
    "-u",
    is_flag=True,
    default=False,
    help=TERM.cyan("Do we wish to upgrade already installed components?"),
)
@click.option(
    "--verbose/--quiet",
    "-v/-q",
    default=False,
    help=TERM.cyan("Do we want verbose output?"),
)
@click.option(
    "--out",
    "-o",
    default=None,
    help=TERM.cyan("Where do you want to output your logs?"),
)
@click.pass_context
def cli(ctx, settings_file, upgrade, verbose, out):
    global log_file, logging_level
    if verbose:
        logging_level = logging.DEBUG
    if out is not None:
        log_file = out
    logging.basicConfig(
        level=logging_level,
        format=log_format,
        handlers=[
            logging.FileHandler(filename=log_file, mode='a', encoding='utf-8'),
            logging.StreamHandler()
        ]
    )
    ctx.obj = Settings(settings_file, upgrade, verbose, out)


@cli.command(help=TERM.cyan("Install Hyperledger Fabric Certificate Authorities"))
@pass_settings
def cert_auth(settings):
    opts = load_config(settings.settings_file)
    runner_ca(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan("Install Hyperledger Composer"))
@pass_settings
def composer(settings):
    opts = load_config(settings.settings_file)
    runner_composer(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan("Upgrade Hyperledger Composer"))
@pass_settings
def composer_up(settings):
    opts = load_config(settings.settings_file)
    runner_composer_up(opts)


@cli.command(help=TERM.cyan("Obtain cryptographic materials from CAs"))
@pass_settings
def crypto(settings):
    opts = load_config(settings.settings_file)
    runner_crypto(opts)


# TODO: Can we compose several CLI commands here to avoid copied code?
@cli.command(help=TERM.cyan("Install end-to-end Fabric/Composer network"))
@pass_settings
def deploy(settings):
    opts = load_config(settings.settings_file)
    runner_deploy(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan("Install end-to-end Hyperledger Fabric network"))
@pass_settings
def fabric(settings):
    opts = load_config(settings.settings_file)
    runner_fabric(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan("Install Hyperledger Fabric Orderers"))
@pass_settings
def orderer(settings):
    opts = load_config(settings.settings_file)
    runner_orderer(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan("Install Hyperledger Fabric Peers"))
@pass_settings
def peer(settings):
    opts = load_config(settings.settings_file)
    runner_peer(opts, upgrade=settings.upgrade)


@cli.command(help=TERM.cyan('Load "nephos" settings YAML file'))
@pass_settings
def settings(settings):
    data = load_config(settings.settings_file)
    logging.info("Settings successfully loaded...\n")
    logging.debug(pretty_print(json.dumps(data, indent=4)))


if __name__ == "__main__":  # pragma: no cover
    cli(obj={})



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import random
from time import sleep

from nephos.fabric.ord import check_ord_tls
from nephos.fabric.settings import get_namespace, get_version
from nephos.fabric.utils import (
    get_helm_pod,
    get_peers,
    get_msps,
    get_channels,
    get_orderers,
    get_an_orderer_msp
)
from nephos.helpers.helm import (
    HelmPreserve,
    helm_check,
    helm_extra_vars,
    helm_install,
    helm_upgrade,
)


def check_peer(namespace, release):
    """Check if Peer is running.

    Args:
        namespace: Namespace where Peer is located.
        release: Name of Peer Helm release.
        

    Returns:
        bool: True once Peer is correctly running.
    """
    pod_exec = get_helm_pod(
        namespace=namespace, release=release, app="hlf-peer"
    )
    res = pod_exec.logs(1000)
    if "Received block" in res:
        return True
    while True:
        if "Starting peer" in res or "Sleeping" in res:
            return True
        else:
            sleep(15)
            res = pod_exec.logs(1000)


# TODO: Split CouchDB creation from Peer creation
def setup_peer(opts, upgrade=False):
    """Setup Peer on K8S.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    for msp in get_msps(opts=opts):
        peer_namespace = get_namespace(opts, msp=msp)
        for release in get_peers(opts=opts, msp=msp):
            # Deploy the CouchDB instances
            version = get_version(opts, "hlf-couchdb")
            config_yaml = f'{opts["core"]["dir_values"]}/{msp}/hlf-couchdb/cdb-{release}.yaml'
            if not upgrade:
                extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
                helm_install(
                    opts["core"]["chart_repo"],
                    "hlf-couchdb",
                    f"cdb-{release}",
                    peer_namespace,
                    extra_vars=extra_vars,

                )
            else:
                preserve = (
                    HelmPreserve(
                        peer_namespace,
                        f"cdb-{release}-hlf-couchdb",
                        "COUCHDB_USERNAME",
                        "couchdbUsername",
                    ),
                    HelmPreserve(
                        peer_namespace,
                        f"cdb-{release}-hlf-couchdb",
                        "COUCHDB_PASSWORD",
                        "couchdbPassword",
                    ),
                )
                extra_vars = helm_extra_vars(
                    version=version, config_yaml=config_yaml, preserve=preserve
                )
                helm_upgrade(
                    opts["core"]["chart_repo"],
                    "hlf-couchdb",
                    f"cdb-{release}",
                    extra_vars=extra_vars,

                )
            helm_check("hlf-couchdb", f"cdb-{release}", peer_namespace)

            # Deploy the HL-Peer charts
            version = get_version(opts, "hlf-peer")
            config_yaml = f"{opts['core']['dir_values']}/{msp}/hlf-peer/{release}.yaml"
            extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
            if not upgrade:
                helm_install(
                    opts["core"]["chart_repo"],
                    "hlf-peer",
                    release,
                    peer_namespace,
                    extra_vars=extra_vars,

                )
            else:
                helm_upgrade(
                    opts["core"]["chart_repo"],
                    "hlf-peer",
                    release,
                    extra_vars=extra_vars,

                )
            helm_check("hlf-peer", release, peer_namespace)
            # Check that peer is running
            check_peer(peer_namespace, release)


def peer_channel_suffix(opts, ord_msp, ord_name):
    """Get command suffix for "peer channel" commands, as they involve speaking with Orderer.

    Args:
        opts (dict): Nephos options dict.
        ord_name (str): Orderer we wish to speak to.
        ord_msp(str): Orderer msp we wish to speark to

    Returns:
        str: Command suffix we need to use in "peer channel" commands.
    """
    ord_tls = check_ord_tls(opts, ord_msp, ord_name)
    if ord_tls:
        cmd_suffix = (
            "--tls "
            + f"--ordererTLSHostnameOverride {ord_name}-hlf-ord "
            + f"--cafile $(ls ${{ORD_TLS_PATH}}/*.pem)"
        )
    else:
        cmd_suffix = ""
    return cmd_suffix


def get_channel_block(peer_ex, ord_name, ord_namespace, channel, cmd_suffix):
    """Get channel block from Peer.

    Args:
        peer_ex (Executor): A Pod Executor representing a Peer.
        ord_name (str): Name of the orderer we wish to communicate with.
        ord_namespace (str): Namespace where the orderer resides.
        channel (str): Name of the channel we with to retrieve.
        cmd_suffix (str): Suffix to the "peer channel fetch" command.

    Returns:
        bool: Were we able to fetch the channel?
    """
    channel_file = f"/var/hyperledger/{channel}.block"
    channel_block, _ = peer_ex.execute(f"ls {channel_file}")
    if not channel_block:
        res, err = peer_ex.execute(
            (
                f"bash -c 'peer channel fetch 0 {channel_file} "
                + f"-c {channel} "
                + f"-o {ord_name}-hlf-ord.{ord_namespace}.svc.cluster.local:7050 {cmd_suffix}'"
            )
        )
        if err:
            return False
    return True


# TODO: Split channel creation from channel joining
def create_channel(opts):
    """Create Channel for Peer.

    Args:
        opts (dict): Nephos options dict.
        
    """
    ord_msp = get_an_orderer_msp(opts=opts)
    ord_namespace = get_namespace(opts, msp=ord_msp)
    ord_name = random.choice(list(get_orderers(opts=opts, msp=ord_msp)))
    cmd_suffix = peer_channel_suffix(opts, ord_msp, ord_name)

    for msp in get_msps(opts=opts):
        peer_namespace = get_namespace(opts, msp=msp)

        for channel in get_channels(opts=opts):
            channel_name = opts["channels"][channel]["channel_name"]
            if msp not in opts["channels"][channel]["msps"]:
                continue
            for index, release in enumerate(get_peers(opts=opts, msp=msp)):
                # Get peer pod
                pod_ex = get_helm_pod(peer_namespace, release, "hlf-peer")

                # Check if the file exists
                has_channel = False
                while not has_channel:
                    has_channel = get_channel_block(
                        pod_ex, ord_name, ord_namespace, channel_name, cmd_suffix
                    )
                    if not has_channel:
                        pod_ex.execute(
                            (
                                "bash -c 'peer channel create "
                                + f"-o {ord_name}-hlf-ord.{ord_namespace}.svc.cluster.local:7050 "
                                + f"-c {channel_name} -f /hl_config/channel/{channel_name}.tx {cmd_suffix}'"
                            )
                        )
                res, _ = pod_ex.execute("peer channel list")
                channels = (res.split("Channels peers has joined: ")[1]).split()
                if channel_name not in channels:
                    pod_ex.execute(
                        (
                            "bash -c "
                            + "'CORE_PEER_MSPCONFIGPATH=$ADMIN_MSP_PATH "
                            + f"peer channel join -b /var/hyperledger/{channel_name}.block {cmd_suffix}'"
                        )
                    )



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from time import sleep

from nephos.fabric.utils import get_helm_pod, get_orderers, get_kafka_configs, get_msps, is_orderer_msp
from nephos.fabric.settings import get_namespace, get_version
from nephos.helpers.helm import helm_check, helm_extra_vars, helm_install, helm_upgrade
from nephos.helpers.misc import execute


def check_ord(namespace, release):
    """Check if Orderer is running.

    Args:
        namespace (str): Namespace where Orderer is located.
        release (str): Name of Orderer Helm release.

    Returns:
        bool: True once Orderer is correctly running.
    """
    pod_exec = get_helm_pod(
        namespace=namespace, release=release, app="hlf-ord"
    )
    res = pod_exec.logs(1000)
    if "fetching metadata for all topics from broker" in res:
        return True
    while True:
        if "Starting orderer" in res:
            return True
        else:
            sleep(15)
            res = pod_exec.logs(1000)


# TODO: We need a similar check to see if Peer uses client TLS as well
def check_ord_tls(opts, ord_msp, ord_name):
    """Check TLS status of Orderer.

    Args:
        opts (dict): Nephos options dict.
        ord_msp (str): orderer msp of the orderer we wish to check the tls
        ord_name (str): orderer name we wish to check the tls
    Returns:
        bool: True if TLS is enabled, False if TLS is disabled.
    """
    ord_namespace = get_namespace(opts, msp=ord_msp)
    ord_tls, _ = execute(
        (
            f"kubectl get cm -n {ord_namespace} "
            + f'{ord_name}-hlf-ord--ord -o jsonpath="{{.data.ORDERER_GENERAL_TLS_ENABLED}}"'
        ),
    )
    return ord_tls == "true"


def setup_ord(opts, upgrade=False):
    """Setup Orderer on K8S.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
    """
    # Kafka
    if "kafka" in opts["ordering"]:
        # Kafka upgrade is risky, so we disallow it by default
        version = get_version(opts, "kafka")
        kafka_config = get_kafka_configs(opts=opts)
        ord_namespace = get_namespace(opts, msp=kafka_config["msp"])
        config_yaml = f"{opts['core']['dir_values']}/{kafka_config['msp']}/kafka/{kafka_config['name']}.yaml"
        extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
        helm_install(
            "incubator",
            "kafka",
            kafka_config['name'],
            ord_namespace,
            extra_vars=extra_vars,
        )
        helm_check(
            "kafka",
            kafka_config['name'],
            ord_namespace,
            pod_num=kafka_config["pod_num"],
        )

    for msp in get_msps(opts=opts):
        if not is_orderer_msp(opts=opts, msp=msp):
            continue
        ord_namespace = get_namespace(opts, msp=msp)
        version = get_version(opts, "hlf-ord")
        for release in get_orderers(opts=opts, msp=msp):
            # HL-Ord
            config_yaml = f'{opts["core"]["dir_values"]}/{msp}/hlf-ord/{release}.yaml'
            extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
            if not upgrade:
                helm_install(
                    opts["core"]["chart_repo"],
                    "hlf-ord",
                    release,
                    ord_namespace,
                    extra_vars=extra_vars,
                )
            else:
                helm_upgrade(
                    opts["core"]["chart_repo"],
                    "hlf-ord",
                    release,
                    extra_vars=extra_vars,
                )
            helm_check("hlf-ord", release, ord_namespace)
            # Check that Orderer is running
            check_ord(ord_namespace, release)



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import OrderedDict
from os import path

import yaml

from nephos.helpers.k8s import context_get
from nephos.fabric.utils import get_msps


# YAML module will load data using an OrderedDict
def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


def dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


yaml.add_representer(OrderedDict, dict_representer)
yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, dict_constructor)


def check_cluster(cluster_name):
    """Check that we are using the correct K8S cluster.

    Args:
        cluster_name (str): Name of K8S cluster.
    """
    context = context_get()
    if context["context"]["cluster"] != cluster_name:
        message = f"We expect to use cluster {cluster_name}, but are instead using cluster {context['context']['cluster']}",
        raise ValueError(message)


def get_namespace(opts, msp=None, ca=None):
    """Get relevant namespace where MSP or CA should be located.

    Args:
        opts (dict): Nephos options dict.
        msp (str): Name of Membership Service Provider (MSP).
        ca (str): Name of Certificate Authority (CA).

    Returns:
        str: Namespace relating to either an MSP or a CA.
    """
    if msp is not None:
        if msp in list(get_msps(opts=opts)):
            if "namespace" in opts["msps"][msp]:
                # Specific MSP-based namespace
                return opts["msps"][msp]["namespace"]
        else:
            raise KeyError(f'Settings dict does not contain MSP "{msp}"')
    elif ca is not None:
        if "cas" in opts and ca in opts["cas"]:
            ca_values = opts["cas"][ca]
        else:
            raise KeyError(f'Settings dict does not contain CA "{ca}"')
        if "namespace" in ca_values:
            # Specific MSP-based namespace
            return ca_values["namespace"]
    # Default case is to return core namespace
    return opts["core"]["namespace"]


def get_version(opts, app):
    """Get version of a specific app

    Args:
        opts (dict): Nephos options dict.
        app (str): Helm application name.

    Returns:
        str: Desired version of Helm app, if specified. Defaults to None.
    """
    if "versions" in opts and app in opts["versions"]:
        return opts["versions"][app]
    else:
        return None


def load_config(settings_file):
    """Load configuration from Nephos options/settings YAML file.

    Args:
        settings_file (str): Name of YAML file containing Nephos options/settings.

    Returns:
        dict: Nephos options/settings.
    """
    with open(settings_file) as f:
        data = yaml.safe_load(f)
    if "cluster" in data["core"]:
        check_cluster(data["core"]["cluster"])
    if path.isdir(data["core"]["chart_repo"]):
        # TODO: This abspath/expanduser combo can be refactored to another function
        data["core"]["chart_repo"] = path.abspath(
            path.expanduser(data["core"]["chart_repo"])
        )
    data["core"]["dir_config"] = path.abspath(
        path.expanduser(data["core"]["dir_config"])
    )
    data["core"]["dir_crypto"] = path.abspath(
        path.expanduser(data["core"]["dir_crypto"])
    )
    data["core"]["dir_values"] = path.abspath(
        path.expanduser(data["core"]["dir_values"])
    )
    return data



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging

from nephos.fabric.ca import setup_ca
from nephos.fabric.crypto import admin_msp, genesis_block, channel_tx, setup_nodes
from nephos.fabric.ord import setup_ord
from nephos.fabric.peer import setup_peer, create_channel
from nephos.composer.install import deploy_composer, install_network, setup_admin
from nephos.composer.upgrade import upgrade_network
from nephos.fabric.utils import get_msps


def runner_ca(opts, upgrade=False):
    """Deploy CAs.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    if opts["cas"]:
        setup_ca(opts, upgrade=upgrade)
    else:
        logging.warning("No CAs defined in Nephos settings, ignoring CA setup")


def runner_composer(opts, upgrade=False):
    """Deploy Hyperledger Composer.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    deploy_composer(opts, upgrade=upgrade)
    setup_admin(opts)
    install_network(opts)


def runner_composer_up(opts):
    """Upgrade Hyperledger Composer network (experimental).

    Args:
        opts (dict): Nephos options dict.
        
    """
    upgrade_network(opts)


def runner_crypto(opts):
    """Create Crypto-material by either using CAs or save Cryptogen material.

    Args:
        opts (dict): Nephos options dict.
        
    """
    # Set up Admin MSPs
    for msp in get_msps(opts=opts):
        admin_msp(opts, msp)
    # Genesis & Channel
    genesis_block(opts)
    # TODO: We currently only support a single channel
    channel_tx(opts)
    # Setup node MSPs
    setup_nodes(opts)


def runner_deploy(opts, upgrade=False):
    """Deploy end-to-end deployment of Hyperledger Fabric and Composer.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    # Fabric
    runner_fabric(opts, upgrade=upgrade)
    # Composer
    runner_composer(opts, upgrade=upgrade)


def runner_fabric(opts, upgrade=False):
    """Deploy Hyperledger Fabric, including CAs/Cryptogen, Orderers and Peers.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    # Setup CA
    runner_ca(opts, upgrade=upgrade)
    # Crypto material
    runner_crypto(opts)
    # Orderers
    runner_orderer(opts, upgrade=upgrade)
    # Peers
    runner_peer(opts, upgrade=upgrade)


def runner_orderer(opts, upgrade=False):
    """Deploy Hyperledger Fabric Orderers.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    setup_ord(opts, upgrade=upgrade)


def runner_peer(opts, upgrade=False):
    """Deploy Hyperledger Fabric Peers.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        
    """
    setup_peer(opts, upgrade=upgrade)
    create_channel(opts)



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from glob import glob
from os import path

from kubernetes.client.rest import ApiException

from nephos.helpers.k8s import Executer, secret_create, secret_from_file, secret_read
from nephos.helpers.misc import execute, rand_string


def credentials_secret(secret_name, namespace, username, password=None):
    """Create a CA credentials secret.

    Args:
        secret_name (str): Name of secret.
        namespace (str): Namespace for secret to be located.
        username (str): Username for credentials secret.
        password (str): Password for credentials secret.

    Returns:
        dict: Secret data including "CA_USERNAME" and "CA_PASSWORD"
    """
    try:
        secret_data = secret_read(secret_name, namespace)
        # Check that the ID stored is the same as Orderer name
        # TODO: Remove asserts here, instead raise error
        assert username == secret_data["CA_USERNAME"]
        if password:
            assert password == secret_data["CA_PASSWORD"]
    except ApiException:
        # Get relevant variables
        if not password:
            password = rand_string(24)
        secret_data = {"CA_USERNAME": username, "CA_PASSWORD": password}
        secret_create(secret_data, secret_name, namespace)
    return secret_data


def crypto_secret(secret_name, namespace, file_path, key):
    """Create a crypto-material secret.

    Args:
        secret_name (str): Name of secret.
        namespace (str): Namespace for secret to be located.
        file_path (str): Path to file we want to store as a secret.
        key (str): Key (file) name of secret we want to store as a secret.
    """
    secret_files = glob(path.join(file_path, "*"))
    if len(secret_files) != 1:
        raise Exception("We should only find one file in this directory")
    secret_from_file(
        secret=secret_name,
        namespace=namespace,
        key=key,
        filename=secret_files[0],
    )


# TODO: Move this to K8S helpers
def get_pod(namespace, identifier, item=0):
    """Get a pod object from K8S.

    Args:
        namespace (str): Namespace where pod is located.
        identifier (str): Name of pod, or a label descriptor.

    Returns:
        Executer: A pod object able to execute commands and return logs.
    """
    node_pod, _ = execute(
        (
            f"kubectl get pods -n {namespace} {identifier} "
            + f'-o jsonpath="{{.items[{item}].metadata.name}}"'
        )
    )
    if not node_pod:
        raise ValueError('"node_pod" should contain a value')
    pod_ex = Executer(node_pod, namespace=namespace)
    return pod_ex


# TODO: Move this to Helm helpers
def get_helm_pod(namespace, release, app, item=0):
    """Get a pod object from K8S.

    Args:
        namespace (str): Namespace where pod is located.
        release (str): Release name of pod.
        app (str): App type of pod.

    Returns:
        Executer: A pod object able to execute commands and return logs.
    """
    identifier = f'-l "app={app},release={release}"'
    return get_pod(namespace, identifier, item=item)


def get_orderers(opts, msp):
    if "orderers" in opts["msps"][msp] and "nodes" in opts["msps"][msp]["orderers"]:
        return opts["msps"][msp]["orderers"]["nodes"].keys()
    return []


# No real purpose in asking the msp for now, but will needed to support multi-org
def get_peers(opts, msp):
    if "peers" in opts["msps"][msp]:
        return opts["msps"][msp]["peers"]["nodes"].keys()
    return []


def is_orderer_msp(opts, msp):
    if "orderers" in opts["msps"][msp] and "nodes" in opts["msps"][msp]["orderers"]:
        return True
    return False


def get_an_orderer_msp(opts):
    for msp in opts["msps"]:
        if is_orderer_msp(opts=opts, msp=msp):
            return msp


def get_msps(opts):
    return opts["msps"].keys()


def get_channels(opts):
    return opts["channels"].keys()


def get_secret_genesis(opts):
    return opts["ordering"]["secret_genesis"]


def get_kafka_configs(opts):
    if "kafka" in opts["ordering"]:
        return opts['ordering']['kafka']



from __future__ import print_function

from collections import namedtuple
from os import path
from time import sleep
import logging

from blessings import Terminal

from nephos.helpers.k8s import pod_check, secret_read
from nephos.helpers.misc import execute

TERM = Terminal()

HelmPreserve = namedtuple(
    "HelmPreserve", ("secret_namespace", "secret_name", "data_item", "values_path")
)
# noinspection PyArgumentList
HelmSet = namedtuple("HelmSet", ("key", "value", "set_string"), defaults=(False,))

CURRENT_DIR = path.abspath(path.split(__file__)[0])


# TODO: We should be able to get the namespace from the Helm release...
def helm_check(app, release, namespace, pod_num=None):
    """Check if a Helm release exists and is functional.

    Args:
        app (str): Helm application name.
        release (str): Release name on K8S.
        namespace (str): Namespace where Helm deployment is located.
        pod_num (int): Number of pods expected to exist in the release.
    """
    identifier = f'-l "app={app},release={release}"'
    pod_check(namespace, identifier, pod_num=pod_num)


def helm_init():
    """Initialise Helm on cluster, using RBAC."""
    res, _ = execute("helm list")
    if res is not None:
        logging.info(TERM.green("Helm is already installed!"))
    else:
        execute(f"kubectl create -f {CURRENT_DIR}/../extras/helm-rbac.yaml")
        execute("helm init --service-account tiller")
        # Fix issue with automountServiceToken
        res, _ = execute(
            "kubectl -n kube-system get deployment tiller-deploy "
            + '-o jsonpath="{.spec.template.spec.automountServiceAccountToken}"'
        )
        if res == "false":
            execute(
                "kubectl -n kube-system patch deployment tiller-deploy "
                + '-p \'{"spec": {"template": {"spec": {"automountServiceAccountToken": true}}}}\''
            )
        # We keep checking the state of helm until everything is running
        running = False
        while not running:
            res, _ = execute("helm list")
            if res is not None:
                running = True
            else:
                print(TERM.red("."), end="", flush=True)
                sleep(15)


def helm_env_vars(env_vars):
    """Convert environmental variables to a "--set" string for Helm deployments.

    Args:
        env_vars (Iterable): Environmental variables we wish to store in Helm.

    Returns:
        str: String containing variables to be set with Helm release.
    """
    env_vars = list(env_vars)
    for i, item in enumerate(env_vars):
        if isinstance(item, tuple):
            item = HelmSet(*item)
        elif not isinstance(item, HelmSet):
            raise TypeError("Items in env_vars array must be HelmSet named tuples")
        env_vars[i] = item
    # Environmental variables
    # TODO: This may well be its own subfunction
    env_vars_string = "".join(
        [
            " --set{} {}={}".format(
                "-string" if item.set_string else "", item.key, item.value
            )
            for item in env_vars
        ]
    )
    return env_vars_string


def helm_preserve(preserve):
    """Convert secret data to a "--set" string for Helm deployments.

    Args:
        preserve (Iterable): Set of secrets we wish to get data from to assign to the Helm Chart.


    Returns:
        str: String containing variables to be set with Helm release.
    """

    env_vars = []
    for item in preserve:
        if isinstance(item, tuple):
            item = HelmPreserve(*item)
        elif not isinstance(item, HelmPreserve):
            raise TypeError("Items in preserve array must be HelmPerserve named tuples")
        secret_data = secret_read(
            item.secret_name, item.secret_namespace
        )
        env_vars.append(HelmSet(item.values_path, secret_data[item.data_item]))
    # Environmental variables
    # TODO: This may well be its own subfunction
    env_vars_string = "".join(
        [
            " --set{} {}={}".format(
                "-string" if item.set_string else "", item.key, item.value
            )
            for item in env_vars
        ]
    )
    return env_vars_string


def helm_extra_vars(
    version=None, config_yaml=None, env_vars=None, preserve=None
):
    """Centralise obtaining extra variables for our helm_install and/or helm_upgrade

    Args:
        version (str): Which Chart version do we wish to install?
        config_yaml (str, Iterable): Values file(s) to override defaults.
        env_vars (Iterable): Environmental variables we wish to store in Helm.
        preserve (Iterable): Set of secrets we wish to get data from to assign to the Helm Chart.


    Returns:
        str: String of Chart version, values files, environmental variables,
    """
    # Get Helm Env-Vars
    extra_vars_string = ""
    if version:
        extra_vars_string += f" --version {version}"
    if config_yaml:
        if isinstance(config_yaml, (str, bytes)):
            config_yaml = (config_yaml,)
        if isinstance(config_yaml, (list, tuple)):
            extra_vars_string += " -f " + " -f ".join(config_yaml)
        else:
            raise ValueError("'config_yaml' variable should be a string, tuple or list")
    if env_vars:
        extra_vars_string += helm_env_vars(env_vars)
    if preserve:
        extra_vars_string += helm_preserve(preserve)
    return extra_vars_string


def helm_install(repo, app, release, namespace, extra_vars=""):
    """Install Helm chart.

    Args:
        repo (str): Repository or folder from which to install Helm chart.
        app (str): Helm application name.
        release (str): Release name on K8S.
        namespace (str): Namespace where to deploy Helm Chart.
        extra_vars (str): Extra variables for Helm including version, values files and environmental variables.

    """
    ls_res, _ = execute(f"helm status {release}")

    if not ls_res:
        command = f"helm install {repo}/{app} -n {release} --namespace {namespace}"
        command += extra_vars
        # Execute
        execute(command)


def helm_upgrade(repo, app, release, extra_vars=""):
    """Upgrade Helm chart.

    Args:
        repo (str): Repository or folder from which to install Helm chart.
        app (str): Helm application name.
        release (str): Release name on K8S.
        extra_vars (str): Extra variables for Helm including version, values files and environmental variables.

    """
    ls_res, _ = execute(f"helm status {release}")

    if ls_res:
        command = f"helm upgrade {release} {repo}/{app}"

        command += extra_vars or ""
        # Execute
        execute(command)
    else:
        raise Exception("Cannot update a Helm release that is not running")



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from kubernetes.client.rest import ApiException

from nephos.composer.connection_template import json_ct
from nephos.fabric.crypto import admin_creds
from nephos.fabric.utils import get_helm_pod
from nephos.fabric.settings import get_namespace, get_version
from nephos.helpers.helm import (
    HelmPreserve,
    helm_check,
    helm_extra_vars,
    helm_install,
    helm_upgrade,
)
from nephos.helpers.k8s import (
    get_app_info,
    cm_create,
    cm_read,
    ingress_read,
    secret_from_file,
)


def get_composer_data(opts, verbose=False):
    """Get Composer deployment data.

    Args:
        opts (dict): Nephos options dict.
        verbose (bool): Verbosity. False by default.

    Returns:
        dict: Data related to the Composer deployment (URI & API key)
    """
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    composer_name = opts["composer"]["name"] + "-hl-composer-rest"
    data = get_app_info(
        peer_namespace,
        composer_name,
        composer_name,
        secret_key="COMPOSER_APIKEY"
    )
    return data


def composer_connection(opts, verbose=False):
    """Composer connection setup.

    This creates a ConfigMap on K8S with the Hyperledger Composer connection.

    Args:
        opts (dict): Nephos options dict.
        verbose (bool): Verbosity. False by default.
    """
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    # TODO: This could be a single function
    peer_msp = opts["peers"]["msp"]
    peer_ca = opts["msps"][peer_msp]["ca"]
    ca_namespace = opts["cas"][peer_ca]["namespace"]
    ingress_urls = ingress_read(
        peer_ca + "-hlf-ca", namespace=ca_namespace
    )
    peer_ca_url = ingress_urls[0]
    try:
        cm_read(opts["composer"]["secret_connection"], peer_namespace)
    except ApiException:
        # Set up connection.json
        # TODO: Improve json_ct to work entirely with opts structure
        cm_data = {
            "connection.json": json_ct(
                opts,
                peer_ca,
                peer_ca_url,
                "AidTech",
                None,
                peer_msp,
                opts["peers"]["channel_name"],
            )
        }
        cm_create(
            cm_data,
            opts["composer"]["secret_connection"],
            peer_namespace,
        )


def deploy_composer(opts, upgrade=False, verbose=False):
    """Deploy Hyperledger Composer on K8S.

    We use the hl-composer Helm chart as a basis to deploying Composer
    on K8S. Please note that Composer is unmaintained and may eventually
    be deprecated from this repository as we migrate to raw Fabric.

    Args:
        opts (dict): Nephos options dict.
        upgrade (bool): Do we upgrade the deployment? False by default.
        verbose (bool): Verbosity. False by default.
    """
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    # Ensure BNA exists
    secret_from_file(
        secret=opts["composer"]["secret_bna"], namespace=peer_namespace
    )
    composer_connection(opts, verbose=verbose)

    # Start Composer
    version = get_version(opts, "hl-composer")
    config_yaml = f'{opts["core"]["dir_values"]}/hl-composer/{opts["composer"]["name"]}.yaml'
    if not upgrade:
        extra_vars = helm_extra_vars(version=version, config_yaml=config_yaml)
        helm_install(
            opts["core"]["chart_repo"],
            "hl-composer",
            opts["composer"]["name"],
            peer_namespace,
            extra_vars=extra_vars
        )
    else:
        preserve = (
            HelmPreserve(
                peer_namespace,
                f"{opts['composer']['name']}-hl-composer-rest",
                "COMPOSER_APIKEY",
                "rest.config.apiKey",
            ),
        )
        extra_vars = helm_extra_vars(
            version=version, config_yaml=config_yaml, preserve=preserve
        )
        helm_upgrade(
            opts["core"]["chart_repo"],
            "hl-composer",
            opts["composer"]["name"],
            extra_vars=extra_vars
        )
    helm_check("hl-composer", opts["composer"]["name"], peer_namespace, pod_num=3)


def setup_card(opts, msp_path, user_name, roles, network=None, verbose=False):
    """Setup the Card for Hyperledger Composer.

    Args:
        opts (dict): Nephos options dict.
        msp_path (str): Path to the MSP on the Composer CLI.
        user_name (str): Name of user for identity card.
        network (str): Name of network for identity card.
        roles (Iterable): Roles to assign to identity card.
        verbose (bool): Verbosity. False by default.
    """

    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    hlc_cli_ex = get_helm_pod(
        peer_namespace, opts["composer"]["name"], "hl-composer"
    )

    # Set up the PeerAdmin card
    ls_res, _ = hlc_cli_ex.execute(
        f"composer card list --card {user_name}@{network}"
    )

    if roles:
        roles_string = "-r " + " -r ".join(roles) + " "
    else:
        roles_string = ""

    if not ls_res:
        hlc_cli_ex.execute(
            (
                "composer card create "
                + (f"-n {network} " if network else "")
                + "-p /hl_config/hlc-connection/connection.json "
                + f"-u {user_name} -c {msp_path}/signcerts/cert.pem "
                + f"-k {msp_path}/keystore/key.pem "
                + f"{roles_string}"
                + f"--file /home/composer/{user_name}@{network}"
            )
        )
        hlc_cli_ex.execute(
            "composer card import "
            + f"--file /home/composer/{user_name}@{network}.card"
        )


def setup_admin(opts, verbose=False):
    """Setup Network admin

    Args:
        opts (dict): Nephos options dict.
        verbose (bool): Verbosity. False by default.
    """
    setup_card(
        opts,
        msp_path="/hl_config/admin",
        user_name="PeerAdmin",
        roles=("PeerAdmin", "ChannelAdmin"),
        verbose=verbose,
    )


def install_network(opts, verbose=False):
    """Install Hyperledger Composer network.

    Args:
        opts (dict): Nephos options dict.
        verbose (bool): Verbosity. False by default.
    """
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    hlc_cli_ex = get_helm_pod(
        peer_namespace, opts["composer"]["name"], "hl-composer"
    )

    # Install network
    # TODO: Getting BNA could be a helper function
    bna, _ = hlc_cli_ex.execute("ls /hl_config/blockchain_network")
    bna_name, bna_rem = bna.split("_")
    bna_version, _ = bna_rem.split(".bna")
    # TODO: This could be a single function
    peer_msp = opts["peers"]["msp"]
    bna_admin = opts["msps"][peer_msp]["org_admin"]
    admin_creds(opts, peer_msp)
    bna_pw = opts["msps"][peer_msp]["org_adminpw"]

    ls_res, _ = hlc_cli_ex.execute(
        f"composer card list --card {bna_admin}@{bna_name}"
    )

    if not ls_res:
        hlc_cli_ex.execute(
            (
                "composer network install --card PeerAdmin@hlfv1 "
                + f"--archiveFile /hl_config/blockchain_network/{bna}"
            )
        )
        hlc_cli_ex.execute(
            (
                "composer network start "
                + "--card PeerAdmin@hlfv1 "
                + f"--networkName {bna_name} --networkVersion {bna_version} "
                + f"--networkAdmin {bna_admin} --networkAdminEnrollSecret {bna_pw}"
            )
        )
        hlc_cli_ex.execute(
            f"composer card import --file {bna_admin}@{bna_name}.card"
        )

    # TODO: This step sometimes fail, maybe needs some try logic to test it,
    # executing it manually after a minute succeeds.
    hlc_cli_ex.execute(
        f"composer network ping --card {bna_admin}@{bna_name}"
    )



#! /usr/bin/env python

#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import logging

from nephos.fabric.settings import get_namespace
from nephos.fabric.utils import get_helm_pod
from nephos.helpers.k8s import secret_from_file

CURRENT_DIR = os.path.abspath(os.path.split(__file__)[0])


def upgrade_network(opts, verbose=False):
    """Upgrade Hyperledger Composer network.

    Args:
        opts (dict): Nephos options dict.
        verbose (bool): Verbosity. False by default.
    """
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    secret_from_file(
        secret=opts["composer"]["secret_bna"], namespace=peer_namespace, verbose=verbose
    )
    # Set up the PeerAdmin card
    hlc_cli_ex = get_helm_pod(peer_namespace, "hlc", "hl-composer", verbose=verbose)

    bna, _ = hlc_cli_ex.execute("ls /hl_config/blockchain_network")
    bna_name, bna_rem = bna.split("_")
    bna_version, _ = bna_rem.split(".bna")
    peer_msp = opts["peers"]["msp"]
    bna_admin = opts["msps"][peer_msp]["org_admin"]

    res, _ = hlc_cli_ex.execute(
        f"composer network ping --card {bna_admin}@{bna_name}"
    )

    curr_version = (res.split("Business network version: ")[1]).split()[0]
    logging.info(curr_version)

    if curr_version != bna_version:
        hlc_cli_ex.execute(
            (
                "composer network install --card PeerAdmin@hlfv1 "
                + f"--archiveFile /hl_config/blockchain_network/{bna}"
            )
        )
        hlc_cli_ex.execute(
            (
                "composer network upgrade "
                + "--card PeerAdmin@hlfv1 "
                + f"--networkName {bna_name} --networkVersion {bna_version}"
            )
        )
        res, _ = hlc_cli_ex.execute(
            f"composer network ping --card {bna_admin}@{bna_name}"
        )
        curr_version = (res.split("Business network version: ")[1]).split()[0]
        logging.info(f"Upgraded to {curr_version}")



from __future__ import print_function

import random
from builtins import input
from getpass import getpass
from os.path import isfile, split
import re
from string import ascii_letters, digits
from subprocess import check_output, STDOUT, CalledProcessError
import time
import logging

from blessings import Terminal
from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import TerminalFormatter

t = Terminal()


def execute(command):
    """Execute an arbitrary command line command.

    Args:
        command (str): Command to execute.
    Returns:
        tuple: 2-tuple of execution info:
        1) result of the command, if successful, None if not;
        2) and error, if command failed, None if not.
    """
    logging.info(t.magenta(command))
    try:
        # TODO: Can we do this with a different command than check_output (Bandit security issue)
        result = check_output(command, stderr=STDOUT, shell=True)
        decoded = result.decode("utf-8")
        logging.debug(decoded)
        return decoded, None
    except CalledProcessError as e:
        error_text = e.output.decode("utf-8")
        logging.error(t.red("Command failed with CalledProcessError:"))
        logging.error(error_text)
        return None, error_text


def execute_until_success(command, delay=15):
    """Execute a command until it is successful.

    Args:
        command (str): Command to execute.
        delay (int): Delay in seconds between each unsuccessful attempt.

    Returns:
        str: result of the command
    """
    res = None
    while not res:
        res, err = execute(
            command,
        )
        if err:
            print(t.red("."), end="", flush=True)
            time.sleep(delay)
        else:
            logging.info(res)
            return res


def input_files(keys, clean_key=False):
    """Read a set of filenames and return data from them.

    Args:
        keys (Iterable): Tuple of keys
        clean_key (bool): Do we clean the key to replace non-alphanumeric symbols with an underscore? False by default.

    Returns:
        dict: Data from each file assigned to its relevant key.
    """
    data = {}
    input_text = "Input {key}"
    for key in keys:
        # TODO: This could be its own function.
        is_file = False
        while not is_file:
            filename = get_response(input_text.format(key=key))
            is_file = isfile(filename)
            if not is_file:
                logging.warning(f"{filename} is not a file")
        if key is None:
            key = split(filename)[1]
            if clean_key:
                dirty_key = key
                key = re.sub(r"[^0-9a-zA-Z_.\-]+", "_", dirty_key)
                if key != dirty_key:
                    logging.warning(t.yellow("Replaced ") + dirty_key + t.yellow(" with ") + key)
        with open(filename, "rb") as f:
            data[key] = f.read()
    return data


def get_response(question, permitted_responses=(), sensitive=False):
    """Get response from user.

    Args:
        question: What do we want to obtain from the user?
        permitted_responses: What responses do we allow?
        sensitive: Is the information sensitive (e.g. a password)?

    Returns:
        str: Response from user.
    """
    logging.info(t.yellow(question))
    if permitted_responses:
        logging.info(t.cyan("Permitted responses: " + str(permitted_responses)))
    responded = 0
    while responded == 0:
        if sensitive:
            response = getpass("Password:")
        else:
            response = input()
        # Check type of response
        if response in permitted_responses or not permitted_responses:
            # Response is among possible responses (or any response is permitted)
            responded = 1
        # Otherwise we ping the user to input a response
        if not responded:
            logging.error(t.red("Invalid response, try again!"))
    return response


def pretty_print(string):
    """Pretty print a JSON string.

    Args:
        string (str): String we want to pretty print.
    """
    return highlight(string, JsonLexer(), TerminalFormatter())


def rand_string(length):
    """Create random alphanumeric string (useful for passwords).

    Args:
        length (int): Length of random string.

    Returns:
        str: Alphanumeric string.
    """
    return "".join(random.choice(ascii_letters + digits) for _ in range(length))



#   Copyright [2018] [Alejandro Vicente Grabovetsky via AID:Tech]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at#
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json

from nephos.fabric.settings import get_namespace

"""Connection template.

This module sets up a connection_json for Hyperledger Composer.
"""


# TODO: We need to improve this to be better organised, and use information from Kubernetes:
# TODO: Peer addresses should depend on whether we are using internal addresses only or using an ingress.
# TODO: Organisation name/subdomain should be a variable
# TODO: Ports should depend on whether internal/external addresses
# TODO: CA Host should be obtained via Kubernetes configuration
def define_orderers(orderer_names, orderer_hosts, domain=None):
    """Define orderers as connection objects.

        Args:
            orderer_names (Iterable): List of orderer names.
            orderer_hosts (Iterable): List of orderer hosts.
            domain (str): Domain used. Defaults to none.

        Returns:
            dict: A dictionary of Orderer Connections
    """
    orderer_connections = {}
    for name, host in zip(orderer_names, orderer_hosts):
        if domain:
            key = f"{name}.{domain}"
        else:
            key = name
        orderer_connections[key] = {"url": ("grpc://" + host + ":7050")}
    return orderer_connections


def define_peers(peer_names, peer_hosts, organisation, domain=None):
    """Define peers as connection objects.

        Args:
            peer_names (Iterable): List of peer names.
            peer_hosts (Iterable): List of peer hosts.
            organisation (str): What organisation the peers belong to
            domain (str): Domain used. Defaults to none.

        Returns:
            tuple: A tuple of dictionaries with Peer Options and Peer Connections.
    """
    peer_options = {}
    peer_connections = {}
    for name, host in zip(peer_names, peer_hosts):
        if domain:
            key = f"{name}.{organisation}.{domain}"
        else:
            key = name
        peer_options[key] = {
            "chaincodeQuery": True,
            "endorsingPeer": True,
            "eventSource": True,
            "ledgerQuery": True,
        }
        peer_connections[key] = {
            "url": ("grpc://" + host + ":7051"),
            "eventUrl": ("grpc://" + host + ":7053"),
        }
    return peer_options, peer_connections


def json_ct(opts, ca_name, ca_host, organisation, domain, msp_id, channel):
    """JSON connection template.

    Args:
        opts (dict): Nephos options dict.
        ca_name (str): Name of CA for Peers.
        ca_host (str): CA host address.
        organisation (str): What organisation the peers belong to.
        domain (str): Domain used.
        msp_id (str): ID of the MSP of the peers.
        channel (str): Channel name.

    Returns:
        dict: A dictionary representing the JSON connection template.
    """
    # Derive variables
    peer_namespace = get_namespace(opts, opts["peers"]["msp"])
    ord_namespace = get_namespace(opts, opts["orderers"]["msp"])
    # TODO: Currently specific to intra-cluster communication (Service)
    peer_hosts = [
        peer + f"-hlf-peer.{peer_namespace}.svc.cluster.local"
        for peer in opts["peers"]["names"]
    ]
    orderer_hosts = [
        orderer + f"-hlf-ord.{ord_namespace}.svc.cluster.local"
        for orderer in opts["orderers"]["names"]
    ]
    # Get peers
    peer_options, peer_connections = define_peers(
        opts["peers"]["names"], peer_hosts, organisation, domain
    )
    peer_names = [key for key, value in peer_options.items()]
    # Get orderers
    orderer_connections = define_orderers(
        opts["orderers"]["names"], orderer_hosts, domain
    )
    orderer_names = [key for key, value in orderer_connections.items()]
    return json.dumps(
        {
            "name": "hlfv1",
            "x-type": "hlfv1",
            "x-commitTimeout": 100,
            "version": "1.0.0",
            "client": {
                "organization": organisation,
                "connection": {
                    "timeout": {
                        "peer": {
                            "endorser": "300",
                            "eventHub": "300",
                            "eventReg": "300",
                        },
                        "orderer": "300",
                    }
                },
            },
            "channels": {channel: {"orderers": orderer_names, "peers": peer_options}},
            "organizations": {
                organisation: {
                    "mspid": msp_id,
                    "peers": peer_names,
                    "certificateAuthorities": [ca_name],
                }
            },
            "orderers": orderer_connections,
            "peers": peer_connections,
            "certificateAuthorities": {
                ca_name: {
                    "url": ("https://" + ca_host + ":443"),
                    "caName": ca_name,
                    # TODO: Ideally this should be set to True
                    "httpOptions": {"verify": False},
                }
            },
        }
    )
